from __future__ import annotations

import functools

import awkward.operations.str as akstr

from dask_awkward.lib.core import Array, map_partitions


def always_highlevel(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        if not kwargs.get("highlevel", True):
            raise ValueError("dask-awkward supports only highlevel awkward arrays.")
        return fn(*args, **kwargs)

    return wrapper


@always_highlevel
def capitalize(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.capitalize,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def center(
    array: Array,
    width,
    padding=" ",
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.center,
        array,
        width=width,
        padding=padding,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def count_substring(
    array: Array,
    pattern: str | bytes,
    *,
    ignore_case: bool = False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.count_substring,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def count_substring_regex(
    array: Array,
    pattern: str | bytes,
    *,
    ignore_case: bool = False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.count_substring_regex,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def ends_with(
    array: Array,
    pattern: str | bytes,
    *,
    ignore_case: bool = False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.ends_with,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def extract_regex(
    array: Array,
    pattern: bytes | str,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.extract_regex,
        array,
        pattern=pattern,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def find_substring(
    array: Array,
    pattern,
    *,
    ignore_case=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.find_substring,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def find_substring_regex(
    array: Array,
    pattern,
    *,
    ignore_case=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.find_substring_regex,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def index_in(
    array: Array,
    value_set,
    *,
    skip_nones=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.index_in,
        array,
        value_set=value_set,
        skip_nones=skip_nones,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_alnum(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_alnum,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_alpha(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_alpha,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_ascii(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_ascii,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_decimal(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_decimal,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_digit(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_digit,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_in(
    array: Array,
    value_set,
    *,
    skip_nones=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_in,
        array,
        value_set=value_set,
        skip_nones=skip_nones,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_lower(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_lower,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_numeric(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_numeric,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_printable(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_printable,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_space(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_space,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_title(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_title,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def is_upper(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.is_upper,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def join(
    array: Array,
    separator,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.join,
        array,
        separator=separator,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def join_element_wise(
    *arrays,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.join_element_wise,
        *arrays,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def length(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.length,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def lower(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.lower,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def lpad(
    array: Array,
    width,
    padding=" ",
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.lpad,
        array,
        width=width,
        padding=padding,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def ltrim(
    array: Array,
    characters,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.ltrim,
        array,
        characters=characters,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def ltrim_whitespace(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.ltrim_whitespace,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def match_like(
    array: Array,
    pattern,
    *,
    ignore_case=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.match_like,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def match_substring(
    array: Array,
    pattern,
    *,
    ignore_case=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.match_substring,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def match_substring_regex(
    array: Array,
    pattern,
    *,
    ignore_case=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.match_substring_regex,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def repeat(
    array: Array,
    num_repeats,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.repeat,
        array,
        num_repeats=num_repeats,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def replace_slice(
    array: Array,
    start,
    stop,
    replacement,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.replace_slice,
        array,
        start=start,
        stop=stop,
        replacement=replacement,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def replace_substring(
    array: Array,
    pattern,
    replacement,
    *,
    max_replacements=None,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.replace_substring,
        array,
        pattern=pattern,
        replacement=replacement,
        max_replacements=max_replacements,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def replace_substring_regex(
    array: Array,
    pattern,
    replacement,
    *,
    max_replacements=None,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.replace_substring_regex,
        array,
        pattern=pattern,
        replacement=replacement,
        max_replacements=max_replacements,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def reverse(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.reverse,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def rpad(
    array: Array,
    width,
    padding=" ",
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.rpad,
        array,
        width=width,
        padding=padding,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def rtrim(
    array: Array,
    characters,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.rtrim,
        array,
        characters=characters,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def rtrim_whitespace(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.rtrim_whitespace,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def slice(
    array: Array,
    start,
    stop=None,
    step=1,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.slice,
        array,
        start=start,
        stop=stop,
        step=step,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def split_pattern(
    array: Array,
    pattern,
    *,
    max_splits=None,
    reverse=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.split_pattern,
        array,
        pattern=pattern,
        max_splits=max_splits,
        reverse=reverse,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def split_pattern_regex(
    array: Array,
    pattern,
    *,
    max_splits=None,
    reverse=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.split_pattern_regex,
        array,
        pattern=pattern,
        max_splits=max_splits,
        reverse=reverse,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def split_whitespace(
    array: Array,
    *,
    max_splits: int | None = None,
    reverse: bool = False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.split_whitespace,
        array,
        max_splits=max_splits,
        reverse=reverse,
        behavior=behavior,
    )


@always_highlevel
def starts_with(
    array: Array,
    pattern,
    *,
    ignore_case=False,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.starts_with,
        array,
        pattern=pattern,
        ignore_case=ignore_case,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def swapcase(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.swapcase,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def title(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.title,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def to_categorical(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.to_categorical,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def trim(
    array: Array,
    characters,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.trim,
        array,
        characters=characters,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def trim_whitespace(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.trim_whitespace,
        array,
        behavior=behavior,
        output_divisions=1,
    )


@always_highlevel
def upper(
    array: Array,
    *,
    highlevel: bool = True,
    behavior: dict | None = None,
) -> Array:
    return map_partitions(
        akstr.upper,
        array,
        behavior=behavior,
        output_divisions=1,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from dask.base import unpack_collections
from dask.highlevelgraph import HighLevelGraph

from dask_awkward.layers import AwkwardInputLayer

if TYPE_CHECKING:
    from dask_awkward.lib.core import Array


def necessary_columns(*args: Any, traverse: bool = True) -> dict[str, list[str]]:
    r"""Determine the columns necessary to compute a collection.

    Parameters
    ----------
    *args : Dask collections or HighLevelGraphs
        The collection (or collection graph) of interest. These can be
        individual objects, lists, sets, or dictionaries.
    traverse : bool, optional
        If True (default), builtin Python collections are traversed
        looking for any Dask collections they might contain.

    Returns
    -------
    dict[str, list[str]]
        Mapping that pairs the input layers in the graph to the
        columns that have been determined necessary from that layer.
        These are not necessarily in the same order as the original
        input.

    Examples
    --------
    If we have a hypothetical parquet dataset (``ds``) with the fields

    - "foo"
    - "bar"
    - "baz"

    And the "baz" field has fields

    - "x"
    - "y"

    The calculation of ``ds.bar + ds.baz.x`` will only require the
    ``bar`` and ``baz.x`` columns from the parquet file.

    >>> import dask_awkward as dak
    >>> ds = dak.from_parquet("some-dataset")
    >>> ds.fields
    ["foo", "bar", "baz"]
    >>> ds.baz.fields
    ["x", "y"]
    >>> x = ds.bar + ds.baz.x
    >>> dak.necessary_columns(x)
    {"from-parquet-abc123": ["bar", "baz.x"]}

    Notice that ``foo`` and ``baz.y`` are not determined to be
    necessary.

    """
    import dask_awkward.lib.optimize as o

    collections, _ = unpack_collections(*args, traverse=traverse)
    if not collections:
        return {}

    out: dict[str, list[str]] = {}
    for obj in collections:
        dsk = obj if isinstance(obj, HighLevelGraph) else obj.dask
        cols_this_dsk = o._necessary_columns(dsk)

        for name in cols_this_dsk:
            neccols = cols_this_dsk[name]
            if not isinstance(dsk.layers[name], AwkwardInputLayer):
                raise TypeError(f"Layer {name} should be an AwkwardInputLayer.")
            cols_this_dsk[name] = o._prune_wildcards(neccols, dsk.layers[name]._meta)

        for key, cols in cols_this_dsk.items():
            prev = out.get(key, [])
            update = list(set(prev + cols))
            out[key] = update

    return out


def sample(arr, factor: int | None = None, probability: float | None = None) -> Array:
    """Decimate the data to a smaller number of rows.

    Must give either `factor` or `probability`.

    Parameters
    ----------
    arr : dask_awkward.Array
        Array collection to sample
    factor : int, optional
        if given, every Nth row will be kept. The counting restarts for each
        partition, so reducing the row count by an exact factor is not guaranteed
    probability : float, optional
        a number between 0 and 1, giving the chance of any particular
        row surviving. For instance, for probability=0.1, roughly 1-in-10
        rows will remain.

    """
    if not (factor is None) ^ (probability is None):
        raise ValueError("Give exactly one of factor or probability")
    if factor:
        return arr.map_partitions(lambda x: x[::factor], meta=arr._meta)
    else:
        return arr.map_partitions(
            lambda x: x[np.random.random(len(x)) < probability], meta=arr._meta
        )
